#!/usr/bin/env python3

"""
PINK create binary test data set filled with random numbers
"""

__author__ = "Bernd Doser"
__email__ = "bernd.doser@h-its.org"
__license__ = "GPLv3"

import argparse
import numpy as np
import random
import os
import struct
import tools

def main():

    parser = argparse.ArgumentParser(description='PINK create binary test data set filled with random numbers')
    parser.add_argument('-o', '--output', default='data.bin', help='Data output file', action=tools.check_extension({'bin'}))
    parser.add_argument('--number-of-images', type=int, default=1, help='Number of images, default = 1')
    parser.add_argument('--width', type=int, default=64, help='Image width (default = 64)')
    parser.add_argument('--height', type=int, default=64, help='Image height (default = 64)')
    parser.add_argument('--depth', type=int, default=1, help='Image depth (default = 1)')
    parser.add_argument('--fill-with', choices=['random', 'zero', 'one', 'diag'], default='random', help='Pixel value (random=default, zero, one)')
    parser.add_argument('--file-version', choices=['1', '2'], default='2', help='File version (1, 2 = default)')
    args = parser.parse_args()

    # <file format version> 0 <data-type> <number of entries> <data layout> <data>
    output = open(args.output, 'wb')

    if args.file_version == '1':
        output.write(struct.pack('i' * 4, args.number_of_images, 1, args.width, args.height))
    elif args.file_version == '2':
        if args.depth == 1:
            output.write(struct.pack('i' * 8, 2, 0, 0, args.number_of_images, 0, 2, args.height, args.width))
        else:
            output.write(struct.pack('i' * 9, 2, 0, 0, args.number_of_images, 0, 3, args.depth, args.height, args.width))
    
    number_of_pixels = args.number_of_images * args.width * args.height * args.depth
    if args.fill_with == 'random':
        floatlist = [random.uniform(0.0, 1.0) for _ in range(number_of_pixels)]
        output.write(struct.pack('%sf' % len(floatlist), *floatlist))
    elif args.fill_with == 'zero':
        floatlist = [0.0 for _ in range(number_of_pixels)]
        output.write(struct.pack('%sf' % len(floatlist), *floatlist))
    elif args.fill_with == 'one':
        floatlist = [1.0 for _ in range(number_of_pixels)]
        output.write(struct.pack('%sf' % len(floatlist), *floatlist))
    elif args.fill_with == 'diag':
        images = np.zeros((args.number_of_images, args.depth, args.height, args.width)).astype(np.float32)
        for i in range(args.height):
            for j in range(args.number_of_images):
                images[j,0,i,i] = 1
        floatlist = images.flatten().tolist()
        output.write(struct.pack('%sf' % len(floatlist), *floatlist))

    print('All done.')

main()



"""
PINK common functions
"""

import argparse
import numpy as np
import math
import os
import struct

def check_extension(choices):
    """ Check in argparse if file extensions are supported """
    
    class Act(argparse.Action):
        def __call__(self,parser,namespace,fname,option_string=None):
            ext = os.path.splitext(fname)[1][1:]
            if ext not in choices:
                option_string = '({})'.format(option_string) if option_string else ''
                parser.error("file doesn't end with one of {}{}".format(choices,option_string))
            else:
                setattr(namespace,self.dest,fname)

    return Act


def ignore_header_comments(inputStream):
    """ Ignore header """

    inputStream.seek(0)
    binary_start_position = 0
    for line in inputStream:
        if line == b'# END OF HEADER\n':
            binary_start_position = inputStream.tell()
            break

    inputStream.seek(binary_start_position)
        
        
def get_header_comments(inputStream):
    """ Return header """

    header = b''
    
    inputStream.seek(0)
    binary_start_position = 0
    for line in inputStream:
        if line == b'# END OF HEADER\n':
            binary_start_position = inputStream.tell()
            break

    inputStream.seek(binary_start_position)
    
    if binary_start_position != 0:
        for line in inputStream:
            header = header + inputStream.readline()
            if line == b'# END OF HEADER\n':
                break   

    inputStream.seek(binary_start_position)     
        
    return header


def calculate_map(somWidth, somHeight, neurons, neuronWidth, neuronHeight, shareIntensity = False, border = 0, shape="box"):
    """ For cartesian map, it reads through the data and creates each neuron as a 1D array
        and then resizes it to the neuronSize print(neurons) """

    if shape == "box":
        neuronSize = np.array([neuronWidth, neuronHeight])
        mapSize = np.array([somWidth, somHeight])
        size = np.multiply(mapSize,np.array(neuronSize) + border) + border
        image = np.empty(size)
        image[:] = np.NAN
        for x in range(somWidth):
            for y in range(somHeight):
                if len(neurons[y + x*somHeight].shape)>0:
                    data = neurons[y + x*somHeight].reshape(neuronWidth, neuronHeight)
                    if not shareIntensity:
                        if np.max(data)-np.min(data) != 0:
                            data = 1.0 * (data - np.min(data)) / (np.max(data) - np.min(data))
                else:
                    data = np.ones((neuronWidth, neuronHeight)) * neurons[y + x*somHeight]
                image[border + x*(neuronWidth + border): (x+1) * (neuronWidth + border), border + y * (neuronHeight + border): (y+1) * (neuronHeight + border)] = data

    #For hexagonal map, it goes through each neuron while accounting for the hexagonal shape, starting at the bottom left corner
    if shape == "hex":
        size = np.multiply((somWidth, somHeight),np.array((neuronWidth, neuronHeight)) + border) + border
        size[1] = math.ceil(size[1] - (somHeight-1.0) * neuronHeight / 4.0)
        size[0] = math.ceil(size[0])
        size = np.asarray(size, int)
        image = np.empty(size)
        image[:] = np.NAN
        mapX = - np.floor((somWidth - 1) / 2)
        mapY = -1
        for neuron in neurons:
            mapY = mapY + 1
            if mapX < 0 and mapY > np.floor((somHeight - 1) / 2):
                mapX = mapX + 1
                mapY = - np.floor((somHeight - 1) / 2) - mapX
            elif mapX >= 0 and mapY > np.floor((somHeight - 1) / 2) - mapX:
                mapX = mapX + 1
                mapY = - np.floor((somHeight - 1) / 2)
            if mapY >= somHeight:
                print("abort")
                return image
            if len(neuron.shape) > 0 and not shareIntensity:
                if np.max(neuron)-np.min(neuron) != 0:
                    neuron = 1.0 * (neuron - np.min(neuron)) / (np.max(neuron) - np.min(neuron))
            for xPos in range(neuronWidth):
                for yPos in range(neuronHeight): 
                    # the corners are ignored
                    if math.floor(xPos/2.0 + yPos) < neuronHeight / 4.0 or \
                       math.floor(xPos/2.0 + yPos + 1) > neuronWidth + neuronHeight / 4.0 or \
                       math.floor(xPos/2.0 - yPos + 1) > neuronHeight / 4.0 or \
                       math.floor(xPos/2.0 - yPos) < -neuronWidth + neuronHeight / 4.0:
                        continue
                    # use the inner hex for plotting
                    else:
                        x = int(math.floor(xPos + ((somWidth - 1) / 2 + mapX + (mapY/2.0))* (neuronHeight+border) + border))
                        y = int(math.floor(yPos + ((somHeight - 1) / 2 + mapY) * (neuronHeight * 3.0/4.0 +border) + border))
                        if len(neuron.shape) > 0:
                            image[x][y] = neuron[yPos + xPos * neuronHeight]
                        else:
                            image[x][y] = neuron
    return image


def show(som, border = 0):
    """ Visualization of PINK SOM """
    
    np_som = np.array(som, copy=False)
    if som.get_som_layout() == "cartesian-2d":
        return calculate_map(np_som.shape[0], np_som.shape[1],
                             np_som.reshape(np_som.shape[0] * np_som.shape[1], np_som.shape[2], np_som.shape[3]),
                             np_som.shape[2], np_som.shape[3], shareIntensity = False, border = border, shape="box")
    if som.get_som_layout() == "hexagonal-2d":
        dim = math.sqrt((4 * np_som.shape[0] - 1) / 3);
        return calculate_map(dim, dim, np_som.reshape(np_som.shape[0], np_som.shape[1] * np_som.shape[2]),
                             np_som.shape[1], np_som.shape[2], shareIntensity = False, border = border, shape="hex")
    else:
        raise RuntimeError("Unsupported SOM layout %s" % som.get_som_layout())


class DataIterator:
    """ Lazy iteration over data file """
    
    def __init__(self, filename):
        self.file = open(filename, 'rb')
        ignore_header_comments(self.file)
        
        # <file format version> 0 <data-type> <number of entries> <data layout> <data>
        version, file_type, data_type, self.number_of_images, layout, dimensionality = struct.unpack('i' * 6, self.file.read(4 * 6))
        self.dimensions = np.fromfile(self.file, count=dimensionality, dtype=np.int32)
        
        print('version:', version)
        print('file_type:', file_type)
        print('data_type:', data_type)
        print('number_of_images:', self.number_of_images)
        print('layout:', layout)
        print('dimensionality:', dimensionality)
        print('dimensions:', self.dimensions)
        print('size:', np.prod(self.dimensions))

        self.data = np.fromfile(self.file, count=np.prod(self.dimensions), dtype=np.float32).reshape(self.dimensions)

    def __iter__(self):
        return self
   
    def __next__(self):
        a = np.fromfile(self.file, count=np.prod(self.dimensions), dtype=np.float32)

        if a.size == np.prod(self.dimensions):
            x = self.data
            self.data = a.reshape(self.dimensions)
            return x
        else:
            raise StopIteration



#!/usr/bin/env python3

import getopt
import numpy
import matplotlib
import struct
import sys
import tools

def print_usage():
    print ('')
    print ('Usage:')
    print ('')
    print ('  showImages.py [Options] <inputfile>')
    print ('')
    print ('Options:')
    print ('')
    print ('  --channel, -c <int>     Number of channel to visualize (default = 0).')
    print ('  --help, -h              Print this lines.')
    print ('  --image, -i <int>       Number of image to visualize (default = 0).')
    print ('  --save, -s <String>     Location to save JPGs to.')
    print ('  --name, -n <String>     Name of saved file.')
    print ('  --display, -d <int>     1 to display SOM as well as save it. Default 0.')
    print ('')

if __name__ == "__main__":

    try:
        opts, args = getopt.getopt(sys.argv[1:],"hi:n:c:n:s:d:",["help", "image=", "channel=", "name=", "save=", "display="])
    except getopt.GetoptError:
        print_usage()
        sys.exit(2)

    imageNumber = 0
    channelNumber = 0
    save = "./"
    name= "image"
    display = 0

    for opt, arg in opts:
        if opt in ("-c", "--channel"):
            channelNumber = int(arg)
        elif opt in ("-h", "--help"):
            print_usage()
            sys.exit()
        elif opt in ("-i", "--image"):
            imageNumber = int(arg)
        elif opt in ("-s", "--save"):
            save = arg
        elif opt in ("-n", "--name"):
            name = arg
        elif opt in ("-d", "--display"):
            display = int(arg)


    if len(args) != 1:
        print_usage()
        print ('ERROR: Input file is missing.')
        sys.exit(1)

    inputfile = args[0]

    print ('Input file is ', inputfile)
    print ('Image number is ', imageNumber)
    print ('Channel number is ', channelNumber)
    
    if(display==1):
        print ('Display is on')
    else:
        print ('Display is off')

    #Importing here allows control of the backend depending on if display is on or off.
    if(display==0):
        matplotlib.use('Agg')
    from matplotlib import pyplot

    file = open(inputfile, 'rb')
    tools.ignore_header_comments(file)
    
    # <file format version> 0 <data-type> <number of entries> <data layout> <data>
    version, file_type, data_type, numberOfImages, layout, dimensionality = struct.unpack('i' * 6, file.read(4 * 6))
    print('version:', version)
    print('file_type:', file_type)
    print('data_type:', data_type)
    print('numberOfImages:', numberOfImages)
    print('layout:', layout)
    print('dimensionality:', dimensionality)

    numberOfChannels = struct.unpack('i', file.read(4))[0] if dimensionality > 2 else 1
    height = struct.unpack('i', file.read(4))[0] if dimensionality > 1 else 1
    width = struct.unpack('i', file.read(4))[0]

    print('width:', width)
    print('height:', height)
    print('numberOfChannels:', numberOfChannels)

    if imageNumber >= numberOfImages:
        print ('Image number too large.')
        sys.exit(1)

    if channelNumber >= numberOfChannels:
        print ('Channel number too large.')
        sys.exit(1)

    size = width * height
    file.seek((imageNumber*numberOfChannels + channelNumber) * size*4, 1)
    array = numpy.array(struct.unpack('f' * size, file.read(size*4)))
    data = numpy.ndarray([width,height], 'float', array)

    fig = pyplot.figure()
    ax = fig.add_subplot(1,1,1)
    ax.set_aspect('equal')
    pyplot.imshow(data, interpolation='nearest', cmap=pyplot.cm.jet)
    pyplot.colorbar()
    fileName = save+'/'+name+'.pdf'
    pyplot.savefig(fileName)
    if(display==1):
        pyplot.show()

    print ('All done.')
    sys.exit()



#!/usr/bin/env python3

import getopt
import numpy
import matplotlib
import struct
import sys
import os.path
import math
import tools

class HeatmapVisualizer():
    def __init__(self, fileName):
        self.__fileName = fileName
        self.__numberOfImages = 0
        self.__somWidth = 0
        self.__somHeight = 0
        self.__somDepth = 0
        self.__maps = []

    def getNumberOfMaps(self):
        return self.__numberOfImages

    def getSomWidth(self):
        return self.__somWidth

    def getSomHeight(self):
        return self.__somHeight

    def getSomDepth(self):
        return self.__somDepth

    def getHeatmap(self, objectNumber):
        return self.__maps[objectNumber]

    #Reads in map data
    def readMap(self):
        #Unpacks the map parameters
        inputStream = open(self.__fileName, 'rb')
        tools.ignore_header_comments(inputStream)
        
        # <file format version> 2 <data-type> <number of entries> <som layout> <data>
        version, file_type, data_type, number_of_data_entries, self.__som_layout, som_dimensionality = struct.unpack('i' * 6, inputStream.read(4 * 6))
        print('version:', version)
        print('file_type:', file_type)
        print('data_type:', data_type)
        print('number_of_data_entries:', number_of_data_entries)
        print('som_layout:', self.__som_layout)
        print('som dimensionality:', som_dimensionality)
        som_dimensions = struct.unpack('i' * som_dimensionality, inputStream.read(4 * som_dimensionality))
        print('som dimensions:', som_dimensions)

        self.__numberOfImages = number_of_data_entries
        self.__somWidth = som_dimensions[0]
        self.__somHeight = som_dimensions[1] if som_dimensionality > 1 else 1
        self.__somDepth = som_dimensions[2] if som_dimensionality > 2 else 1

        print ("images: " + str(self.__numberOfImages))
        print ("width: " + str(self.__somWidth))
        print ("height: " + str(self.__somHeight))
        print ("depth: " + str(self.__somDepth))

        if self.__som_layout == 0: # box
            data = numpy.ones(self.__somWidth * self.__somHeight * self.__somDepth)
            for i in range(self.__somWidth * self.__somHeight * self.__somDepth):
                data[i] = struct.unpack_from("f", inputStream.read(4))[0]
            self.__maps.append(data)
        else: # hex
            data = numpy.ones(hexSize)
            for i in range(hexSize):
                data[i] = struct.unpack_from("f", inputStream.read(4))[0]
            self.__maps.append(data)

        self.__maps = numpy.array(self.__maps)
        print (str(len(self.__maps)) + " maps loaded")

    # Checks if hexagonal or cartesian map is used
    def isHexMap(self):
        return self.__som_layout == 1

    # Creates the SOM and saves it to the specified location. Displays map if --display, -d is set to 1.
    def showMap(self, imageNumber, neuronSize, shareIntensity = False, borderWidth = 2, facecolor = '#ffaadd'):
        print(shareIntensity)
        if facecolor != "#ffaadd":
            print ("WARNING! using non recommended background color! The results will look ugly.")
        figure = pyplot.figure(figsize=(16, 16))
        figure.patch.set_alpha(0.0)
        figure.patch.set_facecolor(facecolor)
        pyplot.subplots_adjust(left = 0.01, right = 0.99, bottom = 0.01, top = 0.99, wspace = 0.1, hspace = 0.1)

        if self.__som_layout == 1:
            print ("hexagonal map")
            image = tools.calculate_map(self.__somWidth, self.__somHeight, self.__maps[imageNumber], neuronSize, neuronSize, shareIntensity=shareIntensity, border=borderWidth, shape="hex")
        else:
            print ("cartesian map")
            image = tools.calculate_map(self.__somWidth, self.__somHeight, self.__maps[imageNumber], neuronSize, neuronSize, shareIntensity=shareIntensity, border=borderWidth, shape="box")

        ax = pyplot.subplot()
        cmap = matplotlib.cm.get_cmap("jet")
        cmap.set_bad(facecolor,1.)
        #Uses image.T because the file is read in as C notation when it's actually Fortran notation.
        ax.imshow(image.T, aspect=1, interpolation="nearest", cmap=cmap)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis("off")
        fileName = save+'/'+name+'%d.pdf' %imageNumber
        pyplot.savefig(fileName,bbox_inches='tight',dpi=150)
        if(display==1):
            pyplot.show()

def print_usage():
    print ('')
    print ('Usage:')
    print ('')
    print ('  showHeatmap.py [Options] <inputfile>')
    print ('')
    print ('Options:')
    print ('')
    print ('  --help, -h              Print this lines.')
    print ('  --image, -i <int>       Number of image to visualize (default = 0).')
    print ('  --border, -b <int>      Border size.')
    print ('  --neuron, -n <int>      Neuron size (default = 12).')
    print ('  --save, -s <String>     Location to save JPGs to.')
    print ('  --name, -o <String>     Name of saved file. Channel number will automatically be added to the end of the name.')
    print ('  --display, -d <int>     1 to display SOM as well as save it. Default 0.')
    print ('')

if __name__ == "__main__":

    try:
        opts, args = getopt.getopt(sys.argv[1:],"h:i:b:n:s:o:d:",["help", "image=", "border=", "neuron=", "save=", "name=", "display="])
    except getopt.GetoptError:
        print_usage()
        sys.exit(1)

    #Default parameters
    save="./"
    name="HEATMAP"
    borderWidth=2
    facecolor="#ffaadd"
    shareIntensity=True
    display=0
    imageNumber = 0
    neuronSize = 12

    #Set parameters
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            print_usage()
            sys.exit()
        elif opt in ("-i", "--image"):
            imageNumber = int(arg)
        elif opt in ("-b", "--border"):
            borderWidth = int(arg)
        elif opt in ("-n", "--neuron"):
            neuronSize = int(arg)
        elif opt in ("-s", "--save"):
            save = arg
        elif opt in ("-o", "--name"):
            name = arg
        elif opt in ("-d", "--display"):
            display = int(arg)
     
    if len(args) != 1:
        print_usage()
        print ('ERROR: Input file is missing.')
        sys.exit(1)

    inputfile = args[0]

    print ('Input file is:', inputfile)
    print ('Save location is:', save)
    print ('The file name is:', name)
    print ('The border width is:', borderWidth)
    print ('The background color is:', facecolor)
    print ('The neuron size is:', neuronSize)
    if(display==1):
        print ('Display is on')
    else:
        print ('Display is off')
     
    #Importing here allows control of the backend depending on if display is on or off.
    if(display==0):
        matplotlib.use('Agg')
    from matplotlib import pyplot
   
    myVisualizer = HeatmapVisualizer(inputfile)
    myVisualizer.readMap()
    myVisualizer.showMap(imageNumber, neuronSize, shareIntensity, borderWidth, facecolor)























    



#!/usr/bin/env python3

"""
PINK test image rotation
"""

__author__ = "Bernd Doser"
__email__ = "bernd.doser@h-its.org"
__license__ = "GPLv3"

import math
import matplotlib.pyplot as plt
import numpy as np
import pink

def main():
    """ Main routine of PINK test image rotation """

    print('PINK version ', pink.__version__)

    images = np.load("../data/all_shapes_shuffled/scaled_images.npy").astype(np.float32)
    
    data = pink.data(images[0, 0, 0:, 0:])

    plt.matshow(images[0, 0, 0:, 0:])
    plt.show()

    print('All done.')

main()



#!/usr/bin/env python3

"""
PINK Training of SOM
"""

__author__ = "Bernd Doser"
__email__ = "bernd.doser@h-its.org"
__license__ = "GPLv3"

import argparse
import math
import matplotlib.pyplot as plt
import numpy as np
import pink
from tqdm import tqdm

class GaussianFunctor():
    """ Returns the value of a Gaussian distribution """

    def __init__(self, sigma=1.1, damping=0.2):
        self.sigma = sigma
        self.damping = damping

    def __call__(self, distance):
        return self.damping / (self.sigma * math.sqrt(2.0 * math.pi)) \
            * math.exp(-0.5 * math.pow((distance / self.sigma), 2))


def main():
    """ Main routine of PINK training """

    print('PINK version ', pink.__version__)

    parser = argparse.ArgumentParser(description='PINK SOM training')
    parser.add_argument('images', nargs='+', help='Input file of images')
    parser.add_argument('-i', '--input-som', help='Input file of SOM initialization')
    parser.add_argument('-o', '--output-som', help='Output file of resulting SOM')
    parser.add_argument('--som-dim', type=int, default=8, help='Dimension of SOM if initialized from scratch')
    parser.add_argument('--neuron-dim', type=int, default=-1, help='Dimension of neurons')
    parser.add_argument('-d', '--display', action='store_true', help='Display SOM during training')
    parser.add_argument('-v', '--verbose', action='store_true', help='Be talkative')
    parser.add_argument('-s', '--scale', action='store_true', help='Scale the input images to be within the range [0, 1]')
    parser.add_argument('-e', '--epochs', type=int, default=1, help='Number of epochs [default: 1]')

    args = parser.parse_args()
    if args.verbose:
        print('Images file:', args.images)
        print('Input SOM file:', args.input_som)
        print('Output SOM file:', args.output_som)
        print('SOM dimension:', args.som_dim)
        print('Display:', args.display)

    images = np.load(args.images[0]).astype(np.float32)
    for image_file in args.images[1:]:
        images = np.append(images, np.load(image_file).astype(np.float32), axis=0)

    # Remove channels
    if len(images.shape) == 4 and images.shape[1] == 1:
        images = np.squeeze(images, axis=1)

    image_dim = images.shape[1]
    neuron_dim = args.neuron_dim if args.neuron_dim != -1 else int(image_dim / math.sqrt(2.0) * 2.0)
    euclid_dim = int(image_dim * math.sqrt(2.0) / 2.0)
    print('Neuron dimension:', neuron_dim)
    print('Euclid dimension:', euclid_dim)

    if args.scale:
        min_element = np.amin(images)
        max_element = np.amax(images)
        factor = 1 / (max_element - min_element)

        print('min value: ', min_element)
        print('max value: ', max_element)
        print('factor: ', factor)

        images = (images - min_element) * factor

    print('min value: ', np.amin(images))
    print('max value: ', np.amax(images))

    # Randomize order of input images
    np.random.shuffle(images)

    if args.verbose:
        print('Image shape: ', images.shape, ', dtype: ', images.dtype)

    if args.input_som:
        # Initialize SOM by input file
        np_som = np.load(args.input_som).astype(np.float32)
    else:
        # Initialize SOM by random values
        np_som = np.random.rand(
            args.som_dim, args.som_dim, neuron_dim, neuron_dim).astype(np.float32)

    if args.verbose:
        print('SOM shape: ', np_som.shape, ', dtype: ', np_som.dtype)

    if args.display:
        new_dim = np_som.shape[0] * np_som.shape[2]
        plt.matshow(np_som.swapaxes(1, 2).reshape((new_dim, new_dim)))
        plt.show()

    som = pink.SOM(np_som)

    trainer = pink.Trainer(som, euclidean_distance_dim=euclid_dim, verbosity=0,
                           distribution_function=pink.GaussianFunctor(sigma=1.1, damping=1.0),
                           number_of_rotations=360, interpolation=pink.Interpolation.BILINEAR,
                           euclidean_distance_type=pink.DataType.UINT8)

    for _ in tqdm(range(args.epochs), desc="epoch"):
        for i in tqdm(range(images.shape[0]), desc="train", leave=False):

            data = pink.Data(images[i])
            trainer(data)

            if args.display and i % 100 == 0:
                trainer.update_som()
                np_som = np.array(som, copy=False)
                new_dim = np_som.shape[0] * np_som.shape[2]
                plt.matshow(np_som.swapaxes(1, 2).reshape((new_dim, new_dim)))
                plt.show()

    if args.output_som:
        trainer.update_som()
        np_som = np.array(som, copy=False)
        np.save(args.output_som, np_som)

    print('All done.')

main()



#!/usr/bin/env python3

import struct
import numpy
import matplotlib
import math
import getopt
import sys
import tools

class MAPVisualizer():
    def __init__(self, fileName):
        self.__fileName = fileName
        self.__numberOfChannels = 1
        self.__somWidth = 1
        self.__somHeight = 1
        self.__somDepth = 1
        self.__neuronWidth = 1
        self.__neuronHeight = 1
        self.__neurons = []

    def getNumberOfChannels(self):
        return self.__numberOfChannels

    def getSomWidth(self):
        return self.__somWidth

    def getSomHeight(self):
        return self.__somHeight

    def getSomDepth(self):
        return self.__somDepth

    #Reads in map data
    def readMap(self):
        #Unpacks the map parameters
        inputStream = open(self.__fileName, 'rb')
        tools.ignore_header_comments(inputStream)
        
        # <file format version> 1 <data-type> <som layout> <neuron layout> <data>
        version, file_type, data_type, som_layout, som_dimensionality = struct.unpack('i' * 5, inputStream.read(4 * 5))
        print('version:', version)
        print('file type:', file_type)
        print('data type:', data_type)
        print('som layout:', som_layout)
        print('som dimensionality:', som_dimensionality)

        self.__somDepth = struct.unpack('i', inputStream.read(4))[0] if som_dimensionality > 2 else 1
        self.__somHeight = struct.unpack('i', inputStream.read(4))[0] if som_dimensionality > 1 else 1
        self.__somWidth = struct.unpack('i', inputStream.read(4))[0]

        print('som depth:', self.__somDepth)
        print('som height:', self.__somHeight)
        print('som width:', self.__somWidth)

        neuron_layout, neuron_dimensionality = struct.unpack('i' * 2, inputStream.read(4 * 2))
        print('neuron layout:', neuron_layout)
        print('neuron dimensionality:', som_dimensionality)

        self.__numberOfChannels = struct.unpack('i', inputStream.read(4))[0] if neuron_dimensionality > 2 else 1
        self.__neuronHeight = struct.unpack('i', inputStream.read(4))[0] if neuron_dimensionality > 1 else 1
        self.__neuronWidth = struct.unpack('i', inputStream.read(4))[0]

        print('neuron depth:', self.__numberOfChannels)
        print('neuron height:', self.__neuronHeight)
        print('neuron width:', self.__neuronWidth)

        #Unpacks data
        try:
            while True:
                data = numpy.ones(self.__neuronWidth * self.__neuronHeight * self.__numberOfChannels)
                for i in range(self.__neuronWidth * self.__neuronHeight * self.__numberOfChannels):
                    data[i] = struct.unpack_from("f", inputStream.read(4))[0]
                self.__neurons.append(data)
        except:
            inputStream.close()
        self.__neurons = numpy.array(self.__neurons)

        print (str(len(self.__neurons)) + " neurons loaded")
        inputStream.close()

    #Checks if hexagonal or cartesian map is used
    def isHexMap(self):
        return len(self.__neurons) < self.__somHeight * self.__somWidth * self.__somDepth

    #Creates the SOM and saves it to the specified location. Displays map if --display, -d is set to 1.
    def showMap(self, channel, shareIntensity = True, borderWidth = 2, facecolor = '#ffaadd'):
        print(shareIntensity)
        if facecolor != "#ffaadd":
            print ("WARNING! using non recommended background color! The results will look ugly.")
        figure = pyplot.figure(figsize=(16, 16))
        figure.patch.set_alpha(0.0)
        figure.patch.set_facecolor(facecolor)
        pyplot.subplots_adjust(left = 0.01, right = 0.99, bottom = 0.01, top = 0.99, wspace = 0.1, hspace = 0.1)
        start=int(len(self.__neurons[0]) / self.__numberOfChannels * channel)
        end=int(len(self.__neurons[0]) / self.__numberOfChannels * (channel+1))
        if self.isHexMap():
            print ("hexagonal map")
            image = tools.calculate_map(self.__somWidth, self.__somHeight, self.__neurons[:,start:end], self.__neuronWidth, self.__neuronHeight, shareIntensity=shareIntensity, border=borderWidth, shape="hex")
        else:
            print ("cartesian map")
            image = tools.calculate_map(self.__somWidth, self.__somHeight, self.__neurons[:,start:end], self.__neuronWidth, self.__neuronHeight, shareIntensity=shareIntensity, border=borderWidth, shape="box")

        ax = pyplot.subplot()
        if somColor==0:
            cmap = matplotlib.cm.get_cmap("jet")
        else:
            if channel==0:
                cmap = matplotlib.cm.get_cmap("Blues")
            elif channel==1:
                cmap = matplotlib.cm.get_cmap("Reds")
            elif channel==2:
                cmap = matplotlib.cm.get_cmap("Greens")
            elif channel==3:
                cmap = matplotlib.cm.get_cmap("Purples")
            elif channel==4:
                cmap = matplotlib.cm.get_cmap("Greys")
            elif channel==5:
                cmap = matplotlib.cm.get_cmap("Oranges")
            else:
                cmap = matplotlib.cm.get_cmap()
        cmap.set_bad(facecolor,1.)
        #Uses image.T because the file is read in as C notation when it's actually Fortran notation.
        ax.imshow(image.T, aspect=1, interpolation="nearest", cmap=cmap)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis("off")
        fileName = save+'/'+name+'%d.pdf' %channel
        pyplot.savefig(fileName,bbox_inches='tight',dpi=150)
        if(display==1):
            pyplot.show()

def print_usage():
    print ('')
    print ('Usage:')
    print ('')
    print ('  showSOM.py [Options] <inputfile>')
    print ('')
    print ('Options:')
    print ('')
    print ('  --help, -h                   Print this lines.')
    print ('  --save, -s <String>          Location to save JPGs to.')
    print ('  --intensity, -i <boolean>    Turn shareIntensity off.')
    print ('  --border, -b <int>           Border size.')
    print ('  --color, -c <String>         Background color. [Defaults to \\#ffaadd - HIGHLY RECOMMENDED!!!]')
    print ('  --name, -n <String>          Name of saved file. Channel number will automatically be added to the end of the name.')
    print ('  --display, -d <int>          1 to display SOM as well as save it. Default 0.')
    print ('  --scolor, -o <int>           0 to use jet color scheme for all SOMs, 1 to use single color scheme with a different color per SOM. Default 1.')
    print ('')

if __name__ == "__main__":

    try:
        opts, args = getopt.getopt(sys.argv[1:],"h:s:b:c:i:n:d:o:",["help", "save=", "border=", "color=", "intensity=", "name=", "display=","scolor="])
    except getopt.GetoptError:
        print_usage()
        sys.exit(1)

    #Default parameters
    save="./"
    name="SOM"
    borderWidth=2
    facecolor="#ffaadd"
    shareIntensity=True
    display=0
    somColor=1

    #Use inputted parameters
    for opt, arg in opts:
        if opt in ("-s", "--save"):
            save = arg
        elif opt in ("-i", "--shareIntensity"):
            shareIntensity = arg in ["True", "true", "yes", "on", "yeah", "si", "wi", "ja"]
        elif opt in ("-c", "--color"):
            facecolor = arg
        elif opt in ("-b", "--border"):
            borderWidth = int(arg)
        elif opt in ("-n", "--name"):
            name = arg
        elif opt in ("-d", "--display"):
            display = int(arg)
        elif opt in ("-o", "--scolor"):
            somColor = int(arg)
        elif opt in ("-h", "--help"):
            print_usage()
            sys.exit()
            
    if len(args) != 1:
        print_usage()
        print ('ERROR: Input file is missing.')
        sys.exit(1)

    inputfile = args[0]

    print ('Input file is:', inputfile)
    print ('Save location is:', save)
    print ('The file name is:', name)
    print ('The border width is:', borderWidth)
    print ('The background color is:', facecolor)
    print ('The intensity is shared:', shareIntensity)
    if(display==1):
        print ('Display is on')
    else:
        print ('Display is off')
     
    #Importing here allows control of the backend depending on if display is on or off.
    if(display==0):
        matplotlib.use('Agg')
    from matplotlib import pyplot
   
    myVisualizer = MAPVisualizer(inputfile)
    myVisualizer.readMap()
    for i in range(0,myVisualizer.getNumberOfChannels()):
        myVisualizer.showMap(i, shareIntensity, borderWidth, facecolor)



#!/usr/bin/env python3

import struct
import numpy
import matplotlib
import math
import getopt
import sys
import tools
from show_heatmap import HeatmapVisualizer as HV

class Rotations():
    def __init__(self, fileName):
        self.__fileName = fileName
        self.__numberOfImages = 0
        self.__somWidth = 0
        self.__somHeight = 0
        self.__somDepth = 0
        self.__flipped = []
        self.__rotations = []

    def getNumberOfImages(self):
        return self.__numberOfImages

    def getSomWidth(self):
        return self.__somWidth

    def getSomHeight(self):
        return self.__somHeight

    def getSomDepth(self):
        return self.__somDepth

    def getImageNeuron(self, objectNumber, neuron):
        return self.__flipped[:,objectNumber * self.__somWidth * self.__somHeight * self.__somDepth + neuron], self.__rotations[:,objectNumber * self.__somWidth * self.__somHeight * self.__somDepth + neuron]

    def getImageBestNeuron(self, objectNumber, mapping):
        mv = HV(mapping)
        mv.readMap()
        bestNeuron = numpy.argmin(mv.getHeatmap(objectNumber))
        return self.getImageNeuron(objectNumber, bestNeuron)

    def getAllImages(self, mapping):
        bestRot = numpy.ones((self.__numberOfImages,2))
        mv = HV(mapping)
        mv.readMap()
        for i in range(self.__numberOfImages):
            bestNeuron = numpy.argmin(mv.getHeatmap(i))
            bestRot[i] = self.getImageNeuron(i, bestNeuron)
        return bestRot

    #Reads in map data
    def readRotations(self):
        #Unpacks the map parameters
        inputStream = open(self.__fileName, 'rb')
        tools.ignore_header_comments(inputStream)
        
        # <file format version> 3 <number of entries> <som layout> <data>
        version, file_type, number_of_data_entries, som_layout, som_dimensionality = struct.unpack('i' * 5, inputStream.read(4 * 5))
        print('version:', version)
        print('file_type:', file_type)
        print('number_of_data_entries:', number_of_data_entries)
        print('som_layout:', som_layout)
        print('som dimensionality:', som_dimensionality)
        som_dimensions = struct.unpack('i' * som_dimensionality, inputStream.read(4 * som_dimensionality))
        print('som dimensions:', som_dimensions)

        self.__numberOfImages = number_of_data_entries
        self.__somWidth = som_dimensions[0]
        self.__somHeight = som_dimensions[1] if som_dimensionality > 1 else 1
        self.__somDepth = som_dimensions[2] if som_dimensionality > 2 else 1

        print ("images: " + str(self.__numberOfImages))
        print ("width: " + str(self.__somWidth))
        print ("height: " + str(self.__somHeight))
        print ("depth: " + str(self.__somDepth))

        #Unpacks data
        try:
            while True:
                dataF = numpy.ones(self.__somWidth * self.__somHeight * self.__somDepth * self.__numberOfImages)
                dataR = numpy.ones(self.__somWidth * self.__somHeight * self.__somDepth * self.__numberOfImages)
                for i in range(self.__somWidth * self.__somHeight * self.__somDepth * self.__numberOfImages):
                    dataF[i] = struct.unpack_from("?", inputStream.read(1))[0]
                    dataR[i] = struct.unpack_from("f", inputStream.read(4))[0]
                self.__flipped.append(dataF)
                self.__rotations.append(dataR)

        except:
            inputStream.close()

        self.__flipped = numpy.array(self.__flipped)
        self.__rotations = numpy.array(self.__rotations)

        print ("rotations loaded")
        inputStream.close()


def print_usage():
    print ('')
    print ('Usage:')
    print ('')
    print ('  rotations.py [Options] <inputfile>')
    print ('')
    print ('Options:')
    print ('')
    print ('  --help, -h                   Print this lines.')
    print ('  --image, -i <int>            Image number.')
    print ('  --mapping, -m <String>       Mapping file for finding best neurons.')
    print ('')

if __name__ == "__main__":

    try:
        opts, args = getopt.getopt(sys.argv[1:],"h:i:m:",["help", "image=", "mapping="])
    except getopt.GetoptError:
        print_usage()
        sys.exit(1)

    # Default parameters
    image = 0
    mapping = ""

    # Use input parameters
    for opt, arg in opts:
        if opt in ("-i", "--image"):
            image = int(arg)
        elif opt in ("-m", "--mapping"):
            mapping = arg
        elif opt in ("-h", "--help"):
            print_usage()
            sys.exit()
            
    if len(args) != 1:
        print_usage()
        print ('ERROR: Input file is missing.')
        sys.exit(1)

    inputfile = args[0]

    print ('Input file is:', inputfile)
    print ('The image number is:', image)
    print ('The mapping file is:', mapping)

    myRotations = Rotations(inputfile)
    myRotations.readRotations()

    if image==-1:
        print(myRotations.getAllImages(mapping))
    else:
        print(myRotations.getImageBestNeuron(image,mapping))



#!/usr/bin/env python3

import getopt
import numpy as np
import os
import struct
import sys

def print_usage():
    print ('')
    print ('Usage:')
    print ('')
    print ('  show_best_rot_flip.py [Options] <inputfile>')
    print ('')
    print ('Options:')
    print ('')
    print ('  --help, -h              Print this lines.')
    print ('  --image, -i <int>       Number of image to visualize (default = 0).')
    print ('')

if __name__ == "__main__":

    try:
        opts, args = getopt.getopt(sys.argv[1:],"h:i:",["help", "image="])
    except getopt.GetoptError:
        print_usage()
        sys.exit(1)

    #Default parameters
    image_number = 0

    #Set parameters
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            print_usage()
            sys.exit()
        elif opt in ("-i", "--image"):
            image_number = int(arg)

    if len(args) != 1:
        print_usage()
        print ('ERROR: Input file is missing.')
        sys.exit(1)

    inputfile = args[0]

    print ('Input file is', inputfile)
    print ('Image number =', image_number)    
    
    filesize = os.path.getsize(inputfile)
    file = open(inputfile, 'rb')
    number_of_images, som_width, som_height, som_depth = struct.unpack('i' * 4, file.read(4*4))

    print ('Number of images =', number_of_images)
    print ('som_width =', som_width)
    print ('som_height =', som_height)
    print ('som_depth =', som_depth)
    
    radius = int((som_width - 1) / 2)
    size_header = 16
    size_pixel = 5
    if (filesize - size_header) / size_pixel / number_of_images == som_width * som_height * som_depth:
        layout = 'cartesian'
        number_of_neurons = som_width * som_height * som_depth
        
    elif (filesize - size_header) / size_pixel / number_of_images == som_width * som_width - radius * (radius + 1):
        layout = 'hexagonal'
        number_of_neurons = som_width * som_width - radius * (radius + 1)
        
    else:
        print ('ERROR: Can\'t detect layout.')
        sys.exit(1)

    print('layout =', layout)
    print('number_of_neurons =', number_of_neurons)

    file.seek(image_number * number_of_neurons * size_pixel, 1)
    data = np.fromfile(file, dtype = np.dtype([('flip', np.int8), ('angle', np.float32)]), count = number_of_neurons)
    print (data)

    print ('All done.')
    sys.exit()



#!/usr/bin/env python3

"""
PINK convert data binary file from version 1 to version 2
"""

__author__ = "Bernd Doser"
__email__ = "bernd.doser@h-its.org"
__license__ = "GPLv3"

import argparse
import os
import struct
import tools

def main():
    """ Main routine of PINK convert data binary file """

    parser = argparse.ArgumentParser(description='PINK convert binary formats')
    parser.add_argument('data', help='Data input file (.bin)', action=tools.check_extension({'bin'}))
    parser.add_argument('-o', '--output', default="out.bin", help='Data output file')
    parser.add_argument('-v', '--verbose', action='store_true', help='Be talkative')
    args = parser.parse_args()

    input = open(args.data, 'rb')
    header = tools.get_header_comments(input)
    
    nb_images, nb_channels, width, height = struct.unpack('i' * 4, input.read(4 * 4))
    size = nb_channels * width * height
    
    # <file format version> 0 <data-type> <number of entries> <data layout> <data>
    output = open(args.output, 'wb')
    output.write(header)

    if nb_channels == 1:
        output.write(struct.pack('i' * 8, 2, 0, 0, nb_images, 0, 2, width, height))
    else:
        output.write(struct.pack('i' * 9, 2, 0, 0, nb_images, 0, 3, width, height, nb_channels))

    for _ in range(nb_images):
        output.write(input.read(size*4))

    print('All done.')

main()



#!/usr/bin/env python3

"""
PINK Training of SOM
"""

__author__ = "Bernd Doser"
__email__ = "bernd.doser@h-its.org"
__license__ = "GPLv3"

import math
import numpy as np
import pink
import tools

class GaussianFunctor():
    """ Returns the value of a Gaussian distribution """

    def __init__(self, sigma=1.1, damping=0.2):
        self.sigma = sigma
        self.damping = damping

    def __call__(self, distance):
        return self.damping / (self.sigma * math.sqrt(2.0 * math.pi)) \
            * math.exp(-0.5 * math.pow((distance / self.sigma), 2))

def main():

    np_data = np.array([[1, 2], [3, 4]], np.float32)
    print(np_data)

    data = pink.data(np_data, data_type="float32", layout="cartesian-2d")

    np_som = np.zeros((2, 2, 2, 2)).astype(np.float32)
    print(np_som)

    som = pink.som(np_som, data_type="float32", som_layout="cartesian-2d", neuron_layout="cartesian-2d")
    
    trainer = pink.trainer(som, GaussianFunctor(sigma=1.1, damping=1.0), number_of_rotations=1, use_flip=False, use_gpu=True)
    trainer(data)
    trainer.update_som()
    
    np_som_res = np.array(som, copy=False)
    print(np_som_res)

    print('All done.')

main()
